use std::cell::RefCell;
use std::rc::Rc;

mod msg;
mod node;

mod crypto;
mod io;
pub mod service;

pub mod prelude {
    pub use crate::msg::Message;
    pub use crate::service::{Service, API as ServiceAPI};
}

use crate as fleck;

use prelude::*;

use io::{FleckIO, IOChannel};

pub struct Fleck {
    io: Rc<FleckIO>,
    services: RefCell<service::ServiceStack>,
    message_registry: msg::MessageRegistry,

    queued_packets: RefCell<Vec<io::Packet>>,
}

impl Fleck {
    pub fn new() -> Rc<Self> {
        let mut res = Rc::new(Self {
            io: Rc::new(FleckIO::default()),
            services: RefCell::new(service::ServiceStack::new()),
            message_registry: msg::MessageRegistry::new(),
            queued_packets: RefCell::new(vec![]),
        });

        res.register_core_services();

        res
    }

    fn register_core_services(&self) {
        let mut svcs = self.services.borrow_mut();
        // Inserted in send-processing order

        // Local discovery
        svcs.give_service(service::core::LocalDiscovery::new());

        // Finally, send the packet
        svcs.give_service(service::core::SendPacket::default());
    }

    pub fn run(&self) {
        self.io.poll(self);
    }

    fn send_queued_packets(&self) {
        let mut svcs = self.services.borrow_mut();
        for mut pkt in self.queued_packets.borrow_mut().drain(..) {
            svcs.process_outgoing(self, &mut pkt);
        }
    }
}

impl ServiceAPI for Fleck {
    fn raw_io(&self) -> &FleckIO {
        &self.io
    }
    fn send_packet(&self, pkt: io::Packet) {
        self.queued_packets.borrow_mut().push(pkt);
    }
}

impl io::IOFeedback for Fleck {
    fn packet(&self, mut packet: io::Packet) {
        let mut svcs = self.services.borrow_mut();
        svcs.process_incoming(self, &mut packet);
        self.send_queued_packets();
    }

    fn minor_tick(&self) {
        let mut svcs = self.services.borrow_mut();
        svcs.process_minor_tick(self);
        self.send_queued_packets();
    }

    fn major_tick(&self) {
        let mut svcs = self.services.borrow_mut();
        svcs.process_major_tick(self);
        self.send_queued_packets();
    }
}
