use std::{
    any::Any,
    hash::{Hash, Hasher},
};

use serde::{de::DeserializeOwned, Deserialize, Serialize};

const MESSAGE_MAGIC: u64 = 0x1234123412341234;

#[derive(Serialize, Deserialize)]
#[repr(u16)]
pub enum HeaderType {
    Hello,
    KeepAlive,
    UserType(u8),
}

#[derive(Serialize, Deserialize)]
pub struct Message {
    ht: HeaderType,
    magic: u64,
}

impl Message {
    pub fn build(ht: HeaderType) -> Self {
        Self {
            ht,
            magic: MESSAGE_MAGIC,
        }
    }
}

pub(crate) struct MessageRegistry {
    deser: std::collections::HashMap<u16, Box<dyn Fn(&[u8]) -> Option<Box<dyn std::any::Any>>>>,
}

pub trait MessageParams {
    const NAME: &'static str;
    const ENCRYPTED: bool;
}

impl MessageRegistry {
    pub(crate) fn new() -> Self {
        Self {
            deser: Default::default(),
        }
    }

    pub(crate) fn add_message_type<MT: MessageParams + Serialize + DeserializeOwned>(&mut self) {
        let mut hasher = std::collections::hash_map::DefaultHasher::new();
        MT::NAME.hash(&mut hasher);
        let derived_typeid = (hasher.finish() & 0xffff) as u16;
        self.deser.insert(
            derived_typeid,
            Box::new(|data| {
                let r = bincode::deserialize(data).ok()?;
                let boxed: Box<dyn Any> = Box::new(r);
                Some(boxed)
            }),
        );
    }

    pub(crate) fn deserialize(&self, data: &[u8]) -> Option<Box<dyn std::any::Any>> {
        let msg_type = u16::from_le_bytes(data[0..1].try_into().unwrap());
        match self.deser.get(&msg_type) {
            Some(f) => f(&data[2..]),
            None => None,
        }
    }
}
