use std::cell::RefCell;
use std::ops::DerefMut;
use std::{borrow::BorrowMut, rc::Rc};

use crate::io::{FleckIO, Packet};

mod lowlevel;
mod priority;

pub use priority::{Never, ServicePriority as Priority};

use self::priority::AbstractServicePriority;

pub mod order {
    pub use super::priority::Never;

    pub struct FirstTag;
    pub type First = super::priority::Fixpoint<FirstTag>;

    pub struct PreprocessingTag;
    pub type Preprocessing = super::priority::Fixpoint<PreprocessingTag>;

    pub struct ProcessingTag;
    pub type Processing = super::priority::Fixpoint<ProcessingTag>;

    pub struct PostprocessingTag;
    pub type Postprocessing = super::priority::Fixpoint<PostprocessingTag>;

    pub struct LastTag;
    pub type Last = super::priority::Fixpoint<FirstTag>;
}

pub(crate) mod core {
    pub(crate) use super::lowlevel::{LocalDiscovery, SendPacket};
}

pub trait API {
    fn raw_io(&self) -> &FleckIO;
    fn send_packet(&self, pkt: Packet);
}

pub trait Service {
    fn process_incoming(&mut self, _api: &dyn API, _packet: &mut Packet) {}
    fn process_outgoing(&mut self, _api: &dyn API, _packet: &mut Packet) {}
    fn process_minor_tick(&mut self, _api: &dyn API) {}
    fn process_major_tick(&mut self, _api: &dyn API) {}
}

pub trait IncomingPriority {
    type Priority: priority::AbstractServicePriority;
}

pub trait OutgoingPriority {
    type Priority: priority::AbstractServicePriority;
}

#[macro_export]
macro_rules! service_priority {
    ($what:ty, in: $i:ty, out: $o:ty) => {
        impl fleck::service::IncomingPriority for $what {
            type Priority = $i;
        }
        impl fleck::service::OutgoingPriority for $what {
            type Priority = $o;
        }
    };
    ($what:ty, in: $i:ty) => {
        impl fleck::service::IncomingPriority for $what {
            type Priority = $i;
        }
        impl fleck::service::OutgoingPriority for $what {
            type Priority = fleck::service::order::Never;
        }
    };
    ($what:ty, out: $o:ty) => {
        impl fleck::service::IncomingPriority for $what {
            type Priority = fleck::service::order::Never;
        }
        impl fleck::service::OutgoingPriority for $what {
            type Priority = $o;
        }
    };
    ($what:ty) => {
        impl fleck::service::IncomingPriority for $what {
            type Priority = fleck::service::order::Never;
        }
        impl fleck::service::OutgoingPriority for $what {
            type Priority = fleck::service::order::Never;
        }
    };
}

#[derive(Default)]
pub struct ServiceStack {
    services: Vec<Rc<RefCell<dyn Service>>>,

    incoming_order: priority::TotalOrder<Rc<RefCell<dyn Service>>>,
    outgoing_order: priority::TotalOrder<Rc<RefCell<dyn Service>>>,

    incoming_cache: Option<Vec<Rc<RefCell<dyn Service>>>>,
    outgoing_cache: Option<Vec<Rc<RefCell<dyn Service>>>>,
}

impl ServiceStack {
    pub(crate) fn new() -> Self {
        Self::default()
        /*Self {
            api: None,
            services: Vec::new(),

            incoming_order: Default::default(),
            outgoing_order: Default::default(),
        }*/
    }

    pub fn give_service<S: 'static + Service + IncomingPriority + OutgoingPriority>(
        &mut self,
        srv: S,
    ) {
        let srv = Rc::new(RefCell::new(srv));

        self.incoming_cache = None;
        self.outgoing_cache = None;

        self.services.push(srv.clone());
        if !<<S as IncomingPriority>::Priority as AbstractServicePriority>::NEVER {
            self.incoming_order
                .add_priority::<<S as IncomingPriority>::Priority>(srv.clone());
        }
        if !<<S as OutgoingPriority>::Priority as AbstractServicePriority>::NEVER {
            self.incoming_order
                .add_priority::<<S as OutgoingPriority>::Priority>(srv.clone());
        }
    }
}

impl Service for ServiceStack {
    fn process_incoming(&mut self, api: &dyn API, packet: &mut Packet) {
        let services = self
            .incoming_cache
            .get_or_insert_with(|| self.incoming_order.clone().order());

        for srv in services {
            srv.as_ref()
                .borrow_mut()
                .deref_mut()
                .process_incoming(api, packet);
        }
    }

    fn process_outgoing(&mut self, api: &dyn API, packet: &mut Packet) {
        let services = self
            .outgoing_cache
            .get_or_insert_with(|| self.outgoing_order.clone().order());

        for srv in services {
            srv.as_ref()
                .borrow_mut()
                .deref_mut()
                .process_incoming(api, packet);
        }
    }

    fn process_minor_tick(&mut self, api: &dyn API) {
        for srv in &self.services {
            srv.as_ref()
                .borrow_mut()
                .deref_mut()
                .process_minor_tick(api);
        }
    }

    fn process_major_tick(&mut self, api: &dyn API) {
        for srv in &self.services {
            srv.as_ref()
                .borrow_mut()
                .deref_mut()
                .process_major_tick(api);
        }
    }
}
/*

    pub(crate) fn process<SP: FnMut(&mut dyn Service, &dyn API) -> ()>(&mut self, mut sp: SP) {
    }

    pub(crate) fn process_reverse<SP: FnMut(&mut dyn Service, &dyn API) -> ()>(
        &mut self,
        mut sp: SP,
    ) {
        let services = self.incoming_cache.get_or_insert_with(|| {
            self.incoming_order.clone().order()
        });

        for srv in self.services.iter_mut().rev() {
            sp(srv.as_ref().borrow_mut().deref_mut(), self.api.as_ref().unwrap().as_ref());
        }
    }
}
*/
